﻿using Microsoft.AspNetCore.Mvc;
using Microsoft.OpenApi.Models;
using Swashbuckle.AspNetCore.SwaggerGen;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;

namespace Mirea.Api.Endpoint.Configuration.SwaggerOptions;

public class ActionResultSchemaFilter : IOperationFilter
{
    public void Apply(OpenApiOperation operation, OperationFilterContext context)
    {
        var returnType = context.MethodInfo.ReturnType;
        if (!returnType.IsEquivalentTo(typeof(ActionResult)) &&
            !returnType.IsEquivalentTo(typeof(ContentResult)) &&
            !returnType.IsEquivalentTo(typeof(FileStreamResult)) &&
            !returnType.IsGenericType)
            return;

        if (returnType.IsGenericType &&
            !returnType.GetGenericTypeDefinition().IsEquivalentTo(typeof(ActionResult<>)) &&
            !returnType.GetGenericTypeDefinition().IsEquivalentTo(typeof(Task<>)))
            return;

        var genericType = returnType.IsGenericType ? returnType.GetGenericArguments().FirstOrDefault() : returnType;
        if (genericType == null)
            return;

        var responseTypeAttributes = context.MethodInfo.GetCustomAttributes(typeof(ProducesResponseTypeAttribute), false)
            .Cast<ProducesResponseTypeAttribute>()
            .Where(attr => attr.StatusCode == 200)
            .ToList();

        var contentType = "application/json";

        if (context.MethodInfo.GetCustomAttributes(typeof(ProducesAttribute), false)
            .FirstOrDefault() is ProducesAttribute producesAttribute)
            contentType = producesAttribute.ContentTypes.FirstOrDefault() ?? "application/json";

        if (responseTypeAttributes.Count != 0)
        {
            var responseType = responseTypeAttributes.First().Type;
            genericType = responseType;
        }

        if (genericType.IsEquivalentTo(typeof(ContentResult)) || genericType.IsEquivalentTo(typeof(FileStreamResult)))
        {
            operation.Responses["200"] = new OpenApiResponse
            {
                Description = "OK",
                Content = new Dictionary<string, OpenApiMediaType>
                {
                    [contentType] = new()
                }
            };
        }
        else if (genericType == typeof(ActionResult))
        {
            operation.Responses["200"] = new OpenApiResponse { Description = "OK" };
        }
        else
        {
            OpenApiSchema schema;
            if (genericType.IsGenericType && genericType.GetGenericTypeDefinition() == typeof(ActionResult<>))
                schema = context.SchemaGenerator.GenerateSchema(genericType.GetGenericArguments().FirstOrDefault(),
                    context.SchemaRepository);
            else
                schema = context.SchemaGenerator.GenerateSchema(genericType, context.SchemaRepository);

            operation.Responses["200"] = new OpenApiResponse
            {
                Description = "OK",
                Content = new Dictionary<string, OpenApiMediaType>
                {
                    [contentType] = new() { Schema = schema }
                }
            };
        }
    }
}